"""Experiment CLI."""

from enum import Enum
from typing import Annotated, Optional

import typer
import yaml
from pydantic import BaseModel, RootModel
from rich.table import Table
from slices_expauth_client.exceptions import ApiException
from slices_expauth_client.models.experiment_member_create import ExperimentMemberCreate
from slices_expauth_client.models.experiment_member_role import ExperimentMemberRole
from slices_expauth_client.models.experiment_member_update import ExperimentMemberUpdate

from . import console, err_console
from .context import Context
from .util.pagination import get_next_page_marker
from .utils import get_error_detail, is_experiment_id, is_type_id

experiment_member_app = typer.Typer(short_help="Commands to list/set/remove experiment members", no_args_is_help=True)


def validate_type_id(value: str, *, required_type: str | None = None):
    """Validate TypeID."""

    if value is None:
        return None

    if not is_type_id(value):
        raise typer.BadParameter("invalid ID")

    type_, domain, suffix = value.split("_")
    if type_ is not None and type_ != required_type:
        raise typer.BadParameter(f"invalid ID type. Expected {required_type!r}, got {type_!r}")

    return value


def validate_user_type_id(value: str):
    """Validate user TypeID."""

    return validate_type_id(value, required_type="user")


def _get_experiment_id(ctx: Context, name: str) -> str:
    if is_experiment_id(name):
        return name
    experiments = ctx.obj.api.experiments_api.get_experiments(
        project_id=ctx.obj.require_project_id(),
        friendly_name=name,
    )

    if not len(experiments.experiments) > 0:
        err_console.print(f"[red]Could not find experiment with name {name} :magnifying_glass_tilted_left:")
        raise typer.Exit(1)

    return experiments.experiments[0].id


ALL_ROLES = [k.lower() for k in ExperimentMemberRole.__members__]
ALL_ALLOWED_ROLES = [k for k in ALL_ROLES if k != "lead"]
ALL_ALLOWED_ROLES_STR = "/".join(ALL_ALLOWED_ROLES)


def validate_role(value: str) -> str:
    """Validate the experiment member role."""

    value = value.lower()
    if value not in ALL_ROLES:
        raise typer.BadParameter(f"Invalid role {value!r}. Allowed roles: {'/'.join(ALL_ALLOWED_ROLES)}")
    if value not in ALL_ALLOWED_ROLES:
        raise typer.BadParameter(
            f"May not change role of user in experiment to {value!r}. Allowed roles: {'/'.join(ALL_ALLOWED_ROLES)}"
        )
    return value


def _map_project_role_to_experiment_role(project_role: str) -> ExperimentMemberRole:
    """Explicit mapping of project roles to experiment roles."""

    # project has member roles: 'lead', 'admin', 'member', 'viewer'
    if project_role.lower() == "lead":
        return ExperimentMemberRole.ADMIN
    if project_role.lower() == "admin":
        return ExperimentMemberRole.ADMIN
    if project_role.lower() == "member":
        return ExperimentMemberRole.MEMBER
    if project_role.lower() == "viewer":
        return ExperimentMemberRole.VIEWER
    raise ValueError(f"Unsupported project role {project_role!r}")


def _try_get_user_name(ctx: Context, user_id: str) -> str:
    try:
        extra_user_info = ctx.obj.api.users_api.get_user(user_id)
    except:  # noqa: E722
        return user_id
    else:
        return f"{extra_user_info.first_name} {extra_user_info.last_name}"


def do_add_missing_project_members_to_experiment(
    ctx: Context, _experiment_name: str, experiment_id: str, *, print_progress: bool = False
):
    """Add all project members that are not in the experiment yet.

    Their role is based on their role in the project.
    """
    current_project_id = ctx.obj.get_current_project_id()

    if not current_project_id:
        err_console.print("[red]No project set. :x:")
        raise typer.Exit

    project_members_response = ctx.obj.api.projects_api.get_project_members(
        current_project_id,
    )
    project_members = project_members_response.project_members

    has_err = False
    for member in project_members:
        if ctx.obj.user_info and member.user_id == ctx.obj.user_info["sub"]:
            # Skip the user itself
            continue
        try:
            _experiment = ctx.obj.api.experiments_api.add_experiment_member(
                experiment_id,
                ExperimentMemberCreate(role=_map_project_role_to_experiment_role(member.role), user_id=member.user_id),
            )
            if print_progress and console.is_terminal:
                console.print(
                    f"[green]:heavy_check_mark: "
                    f"[bold]{_try_get_user_name(ctx, member.user_id)}[/bold] added "
                    f"with role [italic]{member.role.upper()}[/italic]",
                )
        except ApiException as err:
            if err.status == 409:  # noqa PLR2004: HTTP_409_CONFLICT
                if print_progress and console.is_terminal:
                    console.print(
                        f":heavy_check_mark: [bold]{_try_get_user_name(ctx, member.user_id)}[/bold] "
                        f"was already a member"
                    )
            else:
                err_console.print(f"[bold red]Could not add experiment member {member.user_id} :x:")
                err_console.print(f"[red]{get_error_detail(err.body)}")
                has_err = True

    if has_err:
        raise typer.Exit(1)


def upsert_experiment_member(ctx: Context, _experiment_name: str, experiment_id: str, user_id: str, role: str):
    """Add or update an experiment member."""

    # experiment_members_page = ctx.obj.api.experiments_api.get_experiment_members(
    #     experiment_id,
    # )
    try:
        added_experiment_member = ctx.obj.api.experiments_api.add_experiment_member(
            experiment_id,
            ExperimentMemberCreate(role=ExperimentMemberRole(role), user_id=user_id),
        )
        if console.is_terminal:
            console.print(
                f"[green]:heavy_check_mark: [bold]{_try_get_user_name(ctx, user_id)}[/bold] "
                f"added with role [italic]{added_experiment_member.role.name}[/italic] ",
            )
    except ApiException as err:
        if err.status == 409:  # noqa PLR2004: HTTP_409_CONFLICT
            try:
                updated_experiment_member = ctx.obj.api.experiments_api.update_experiment_member(
                    user_id,
                    experiment_id,
                    ExperimentMemberUpdate(role=ExperimentMemberRole(role)),
                )
                if console.is_terminal:
                    console.print(
                        f"[green]:heavy_check_mark: [bold]{_try_get_user_name(ctx, user_id)}[/bold] "
                        f"updated to role [italic]{updated_experiment_member.role.name}[/italic] ",
                    )
            except ApiException as err2:
                err_console.print("[bold red]Could not update experiment member :x:")
                err_console.print(f"[red]{get_error_detail(err2.body)}")
                raise typer.Exit(1) from err2
        else:
            err_console.print("[bold red]Could not add experiment member :x:")
            err_console.print(f"[red]{get_error_detail(err.body)}")
            raise typer.Exit(1) from err


@experiment_member_app.command(name="set", rich_help_panel="Commands")
def set_experiment_member(
    ctx: Context,
    experiment_name: Annotated[str, typer.Argument(help="Name or ID of the experiment.")],
    user_id: Annotated[str, typer.Argument(help="ID of the user", callback=validate_user_type_id)],
    role: Annotated[
        str,
        typer.Argument(help=f"Role of the user ({ALL_ALLOWED_ROLES_STR})", callback=validate_role),
    ] = "member",
):
    """Add or update an experiment member."""

    experiment_id = _get_experiment_id(ctx, experiment_name)
    upsert_experiment_member(ctx, experiment_name, experiment_id, user_id, role)


@experiment_member_app.command(name="add-project", rich_help_panel="Commands")
def add_project_members_to_experiment_member(
    ctx: Context,
    experiment_name: Annotated[str, typer.Argument(help="Name or ID of the experiment.")],
):
    """Add project members to the experiment (if not yet added).

    This does not change the role of existing project members in the experiment.
    The role in the experiment is based on the role in the project.
    """

    experiment_id = _get_experiment_id(ctx, experiment_name)
    do_add_missing_project_members_to_experiment(ctx, experiment_name, experiment_id, print_progress=True)


@experiment_member_app.command(name="transfer-lead", rich_help_panel="Advanced")
def transfer_experiment_lead(
    ctx: Context,
    experiment_name: Annotated[str, typer.Argument(help="Name or ID of the experiment.")],
    user_id: Annotated[str, typer.Argument(help="ID of the new LEAD user", callback=validate_user_type_id)],
):
    """Transfer the LEAD of the experiment from you to another user."""

    experiment_id = _get_experiment_id(ctx, experiment_name)

    try:
        _experiment = ctx.obj.api.experiments_api.update_experiment_member(
            user_id,
            experiment_id,
            ExperimentMemberUpdate(role=ExperimentMemberRole(ExperimentMemberRole.LEAD)),
        )
        if console.is_terminal:
            console.print(
                f"[green]:heavy_check_mark: Transferred experiment {experiment_name!r} LEAD to [bold]{user_id}[/bold] "
            )
    except ApiException as err:
        err_console.print("[bold red]Could not transfer experiment LEAD (are you the LEAD?) :x:")
        err_console.print(f"[red]{get_error_detail(err.body)}")
        raise typer.Exit(1) from err


class ListFormat(Enum):
    """List format options."""

    text = "text"
    json = "json"
    yaml = "yaml"


class EnrichedExperimentMember(BaseModel):
    """Experiment Member with extra user info."""

    experiment_id: str
    role: ExperimentMemberRole
    user_id: str
    first_name: Optional[str]
    last_name: Optional[str]
    email: Optional[str]

    @property
    def name(self) -> Optional[str]:
        """Name of the user."""
        if self.first_name and self.last_name:
            return f"{self.first_name} {self.last_name}"
        return self.first_name or self.last_name


ExperimentsMembers = RootModel[list[EnrichedExperimentMember]]


@experiment_member_app.command(name="list", rich_help_panel="Commands")
def list_experiment_members(
    ctx: Context,
    experiment_name: Annotated[str, typer.Argument(help="Name or ID of the experiment.")],
    show_ids: Annotated[bool, typer.Option("--show-ids/--no-show-ids", "-I", help="Show internal user IDs.")] = True,
    format: Annotated[ListFormat, typer.Option("--format", "-f")] = "text",  # type: ignore
):
    """List experiment members."""

    experiment_id = _get_experiment_id(ctx, experiment_name)

    experiment_members_page = ctx.obj.api.experiments_api.get_experiment_members(experiment_id)
    all_experiment_members = []
    while experiment_members_page:
        all_experiment_members.extend(experiment_members_page.experiment_members)

        marker = get_next_page_marker(experiment_members_page.experiment_member_links)
        experiment_members_page = (
            ctx.obj.api.experiments_api.get_experiment_members(experiment_id, marker=marker) if marker else None
        )

    # enrich user info
    all_experiment_members_extra = []
    for m in all_experiment_members:
        try:
            extra_info = ctx.obj.api.users_api.get_user(m.user_id)
        except Exception as e:  # noqa: E722
            err_console.print(f"Warning: problem fetching extra user info: {e}")
            extra_info = None
        all_experiment_members_extra.append(
            EnrichedExperimentMember(
                **m.model_dump(),
                first_name=extra_info.first_name if extra_info else None,
                last_name=extra_info.last_name if extra_info else None,
                email=extra_info.email if extra_info else None,
            )
        )

    if format == ListFormat.json:
        console.print_json(
            ExperimentsMembers(all_experiment_members_extra).model_dump_json(by_alias=True, exclude_unset=True)
        )
        return
    if format == ListFormat.yaml:
        console.print(
            yaml.safe_dump(
                ExperimentsMembers(all_experiment_members_extra).model_dump(
                    mode="json", by_alias=True, exclude_unset=True
                )
            )
        )
        return

    if show_ids:
        table = Table("User ID", "Name", "Role", title=f"Members of experiment {experiment_name}")
        for member in all_experiment_members_extra:
            table.add_row(member.user_id, member.name, member.role.upper())
    else:
        table = Table("User", "Role", title=f"Members of experiment {experiment_name}")
        for member in all_experiment_members_extra:
            table.add_row(member.name, member.role.upper())

    console.print(table)


@experiment_member_app.command("remove", rich_help_panel="Commands")
def remove_experiment_member(
    ctx: Context,
    experiment_name: Annotated[str, typer.Argument(..., help="Name or ID of experiment.")],
    user_id: Annotated[
        str, typer.Argument(..., help="User ID to remove from experiment.", callback=validate_user_type_id)
    ],
):
    """Remove an experiment member."""

    experiment_id = _get_experiment_id(ctx, experiment_name)

    try:
        deleted_experiment_member = ctx.obj.api.experiments_api.delete_experiment_member(user_id, experiment_id)
    except ApiException as err:
        err_console.print("[bold red]Could not remove experiment member:x:")
        err_console.print(f"[red]{get_error_detail(err.body)}")
        raise typer.Exit(1) from err

    console.print(
        f":wastebasket: Removed experiment member '{_try_get_user_name(ctx, deleted_experiment_member.user_id)}' "
        f"from experiment {experiment_name}."
    )
